//
//  WHCServieRouter+Router.m
//  WHCServieRouter
//
//  Created by yangke on 2017/9/15.
//  Copyright © 2017年 jackie@youzan. All rights reserved.
//

#import "WHCServieRouter+Router.h"

#define WSLog(msg) NSLog(@"[WHCServieRouter] %@", (msg))
#define WSKey(URL) [WHCServieRouter keyForURL:URL]

NSString *const kWHCServieRouterURL = @"KWHCServieRouteURL";
NSString *const kWHCServieRouterCompletion = @"kWHCServieRouterCompletion";

@implementation WHCServieRouter (Router)

+ (nonnull NSString*)keyForURL:(nonnull NSString*)urlStr {
    NSURL *URL = [NSURL URLWithString:urlStr];
    NSString *key = [NSString stringWithFormat:@"%@%@", URL.host, URL.path];
    return key;
}

+ (nullable NSDictionary*)parametersInURL:(nonnull NSString*)urlStr {
    NSURL *URL = [NSURL URLWithString:urlStr];
    NSMutableDictionary *params = nil;
    NSString *query = URL.query;
    if(query.length > 0) {
        params = [NSMutableDictionary dictionary];
        NSArray *list = [query componentsSeparatedByString:@"&"];
        for (NSString *param in list) {
            NSArray *elts = [param componentsSeparatedByString:@"="];
            if([elts count] < 2) continue;
            NSString *decodedStr = [[elts lastObject] stringByRemovingPercentEncoding];
            [params setObject:decodedStr forKey:[elts firstObject]];
        }
    }
    return params;
}

+ (NSMutableDictionary*)routes {
    @synchronized (self) {
        static NSMutableDictionary *_routes = nil;
        if (!_routes) {
            _routes = [NSMutableDictionary dictionary];
        }
        return _routes;
    }
}

+ (void)bindURL:(nonnull NSString *)urlStr toHandler:(nonnull WHCServieRouterHandler)handler {
    [self.routes setObject:handler forKey:WSKey(urlStr)];
}

+ (void)unbindURL:(nonnull NSString *)urlStr {
    [self.routes removeObjectForKey:WSKey(urlStr)];
}

+ (void)unbindAllURLs {
    [self.routes removeAllObjects];
}

+ (nullable WHCServieRouterHandler)handlerForURL:(nonnull NSString *)urlStr {
    return [self.routes objectForKey:WSKey(urlStr)];
}

+ (BOOL)canHandleURL:(nonnull NSString *)urlStr {
    if (urlStr.length == 0) {
        return NO;
    }
    if ([self handlerForURL:urlStr]) {
        return YES;
    } else {
        return NO;
    }
}

+ (nullable id)handleURL:(nonnull NSString *)urlStr {
    return [self handleURL:urlStr complexParams:nil completion:nil];
}

+ (nullable id)handleURL:(nonnull NSString *)urlStr
              completion:(nullable WHCServieRouterCompletion)completion {
    return [self handleURL:urlStr complexParams:nil completion:completion];
}

+ (nullable id)handleURL:(nonnull NSString *)urlStr
           complexParams:(nullable NSDictionary*)complexParams
              completion:(nullable WHCServieRouterCompletion)completion {
    id obj = nil;
    @try {
        WHCServieRouterHandler handler = [self handlerForURL:urlStr];
        NSMutableDictionary *params = [NSMutableDictionary dictionaryWithDictionary:complexParams];
        [params addEntriesFromDictionary:[self.class parametersInURL:urlStr]];
        [params setObject:urlStr forKey:kWHCServieRouterURL];
        if (completion) {
            [params setObject:completion forKey:kWHCServieRouterCompletion];
        }
        if (!handler) {
            NSString *reason = [NSString stringWithFormat:@"Cannot find handler for route url %@", urlStr];
            NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];
            [userInfo setValue:@(WSExceptionUrlHandlerNotFound) forKey:kWHCServieRouterExceptionCode];
            [userInfo setValue:urlStr forKey:kWHCServieRouterExceptionURLStr];
            [userInfo setValue:params forKey:kWHCServieRouterExceptionURLParams];
            NSException *exception = [[NSException alloc] initWithName:WHCServieRouterExceptionName                                                                reason:reason
                                                              userInfo:userInfo];
            WHCServieRouterExceptionHandler handler = [self getExceptionHandler];
            if (handler) {
                obj = handler(exception);
            }
            WSLog(reason);
        } else {
            obj = handler(params);
        }
    } @catch (NSException *exception) {
        NSMutableDictionary *userInfo = [NSMutableDictionary dictionaryWithDictionary:exception.userInfo];
        [userInfo setValue:@(WSExceptionDefaultCode) forKey:kWHCServieRouterExceptionCode];
        [userInfo setValue:urlStr forKey:kWHCServieRouterExceptionURLStr];
        [userInfo setValue:complexParams forKey:kWHCServieRouterExceptionURLParams];
        NSException *ex = [[NSException alloc] initWithName:exception.name
                                                     reason:exception.reason
                                                   userInfo:userInfo];
        WHCServieRouterExceptionHandler handler = [self getExceptionHandler];
        if (handler) {
            obj = handler(ex);
        }
        WSLog(exception.reason);
    } @finally {
        return obj;
    }
}

+ (void)completeWithParameters:(nullable NSDictionary*)params result:(_Nullable id)result {
    WHCServieRouterCompletion completion = params[kWHCServieRouterCompletion];
    if (completion) {
        completion(result);
    }
}

@end
