//
//  Bifrost.h
//  Bifrost
//
//  Created by yangke on 2017/9/15.
//  Copyright © 2017年 jackie@youzan. All rights reserved.
//

#import <Foundation/Foundation.h>
#import "WHCServieRouterProtocol.h"

#define WSRegister(service_protocol) [WHCServieRouter registerService:@protocol(service_protocol) withModule:self.class];
#define WSModule(service_protocol) ((id<service_protocol>)[WHCServieRouter moduleByService:@protocol(service_protocol)])
#define WSStr(fmt, ...) [NSString stringWithFormat:fmt, ##__VA_ARGS__]

typedef NS_ENUM(NSInteger, WHCServieRouterExceptionCode)
{
    WSExceptionDefaultCode = -20001,
    WSExceptionUrlHandlerNotFound = -20002,
    WSExceptionModuleNotFoundException = -20003,
    WSExceptionAPINotFoundException = -20004,
    WSExceptionFailedToRegisterModule = -20005,
    WSExceptionFailedToSetupModule = -20006,
    WSExceptionFailedToFindModuleByService = -20007,

};
// BifrostException exception name
extern NSExceptionName _Nonnull WHCServieRouterExceptionName;
// Bifrost Exception userInfo keys
extern NSString *const _Nonnull kWHCServieRouterExceptionCode;
extern NSString *const _Nonnull kWHCServieRouterExceptionURLStr;
extern NSString *const _Nonnull kWHCServieRouterExceptionURLParams;
extern NSString *const _Nonnull kWHCServieRouterExceptionServiceProtocolStr;
extern NSString *const _Nonnull kWHCServieRouterExceptionModuleClassStr;
extern NSString *const _Nonnull kWHCServieRouterExceptionAPIStr;
extern NSString *const _Nonnull kWHCServieRouterExceptionAPIArguments;

@interface NSException (WHCServieRouter)
- (WHCServieRouterExceptionCode)bf_exceptionCode;
@end

/**
 The handler for exceptions, like url not found, api not support, ...

 @param exception exceptions when handling route URLs or module APIs
 @return The substitute return object
 */
typedef _Nullable id (^WHCServieRouterExceptionHandler)(NSException * _Nonnull exception);

@interface WHCServieRouter : NSObject

/**
 Method to set exception handler

 @param handler the handler block
 */
+ (void)setExceptionHandler:(WHCServieRouterExceptionHandler _Nullable )handler;

+ (WHCServieRouterExceptionHandler _Nullable )getExceptionHandler;

/**
 Method to register the module srevice with module class.
 Each Module do the registeration before app launch event, like in the +load method.

 @param serviceProtocol the protocol for the module's service
 @param moduleClass The class of the module
 */
+ (void)registerService:(Protocol*_Nonnull)serviceProtocol
             withModule:(Class<WHCServieRouterModuleProtocol> _Nonnull)moduleClass;

/**
 Method to unregister service

 @param serviceProtocol the protocol for the module's service
 */
+ (void)unregisterService:(Protocol*_Nonnull)serviceProtocol;

/**
 Method to setup all registered modules.
 It's recommended to invoke this method in AppDelegate's willFinishLaunchingWithOptions method.
 */
+ (void)setupAllModules;

/**
 Get module instance by service protocol. 
 It's recomended to use macro WSModule for convenient

 @param serviceProtocol the service protocol used to register the module
 @return module instance
 */
+ (id<WHCServieRouterModuleProtocol> _Nullable)moduleByService:(Protocol*_Nonnull)serviceProtocol;

//+ (NSArray<Protocol*>*_Nonnull)allRegisteredServices;
//

/**
 Method to get all registered module classes, sorted by module priority.

 @return module class array, not module instances
 */
+ (NSArray<Class<WHCServieRouterModuleProtocol>>*_Nonnull)allRegisteredModules;

/**
 Method to enumarate all modules for methods in UIApplicationDelegate.
 
 @param selector app delegate selector
 @param arguments argument array
 @return the return value of the method implementation in those modules
 */
+ (BOOL)checkAllModulesWithSelector:(nonnull SEL)selector
                          arguments:(nullable NSArray*)arguments;

@end
