//
//  WHCServieRouter.m
//  WHCServieRouter
//
//  Created by yangke on 2017/9/15.
//  Copyright © 2017年 jackie@youzan. All rights reserved.
//

#import "WHCServieRouter.h"
#import <objc/message.h>
#import <objc/runtime.h>

#define WSLog(msg) NSLog(@"[WHCServieRouter] %@", (msg))
#define WSInstance [WHCServieRouter sharedInstance]

NSExceptionName WHCServieRouterExceptionName = @"WHCServieRouterExceptionName";
NSString * const kWHCServieRouterExceptionCode = @"kWHCServieRouterExceptionCode";
NSString * const kWHCServieRouterExceptionURLStr = @"kWHCServieRouterExceptionURLStr";
NSString * const kWHCServieRouterExceptionURLParams = @"kWHCServieRouterExceptionURLParams";
NSString * const kWHCServieRouterExceptionServiceProtocolStr = @"kWHCServieRouterExceptionServiceProtocolStr";
NSString * const kWHCServieRouterExceptionModuleClassStr = @"kWHCServieRouterExceptionModuleClassStr";
NSString * const kWHCServieRouterExceptionAPIStr = @"kWHCServieRouterExceptionAPIStr";
NSString * const kWHCServieRouterExceptionAPIArguments = @"kWHCServieRouterExceptionAPIArguments";

@implementation NSException (WHCServieRouter)
- (WHCServieRouterExceptionCode)bf_exceptionCode {
    return [self.userInfo[kWHCServieRouterExceptionCode] integerValue];
}
@end

@interface NSObject (WHCServieRouter)
- (void)bf_doesNotRecognizeSelector:(SEL)aSelector;
@end

@interface WHCServieRouter() {
    
}
@property (nonatomic, copy) WHCServieRouterExceptionHandler _Nullable exceptionHandler;
@property (nonatomic, strong) NSMutableDictionary *moduleDict; // <moduleName, moduleClass>
@property (nonatomic, strong) NSMutableDictionary *moduleInvokeDict;
+ (instancetype _Nonnull )sharedInstance;
@end

@implementation WHCServieRouter

+ (instancetype _Nonnull )sharedInstance
{
    static WHCServieRouter *instance = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        instance = [[self alloc] init];
        instance.moduleDict = [NSMutableDictionary dictionary];
        instance.moduleInvokeDict = [NSMutableDictionary dictionary];
    });
    return instance;
}

+ (void)setExceptionHandler:(WHCServieRouterExceptionHandler _Nullable )handler {
    WSInstance.exceptionHandler = handler;
}

+ (WHCServieRouterExceptionHandler _Nullable )getExceptionHandler {
    return WSInstance.exceptionHandler;
}

+ (void)registerService:(Protocol*_Nonnull)serviceProtocol
             withModule:(Class<WHCServieRouterModuleProtocol> _Nonnull)moduleClass {
    NSString *protocolStr = NSStringFromProtocol(serviceProtocol);
    NSString *moduleStr = NSStringFromClass(moduleClass);
    Class class = moduleClass; // to avoid warning
    NSString *exReason = nil;
    if (protocolStr.length == 0) {
        exReason =  WSStr(@"Needs a valid protocol for module %@", moduleStr);
    } else if (moduleStr.length == 0) {
        exReason =  WSStr(@"Needs a valid module for protocol %@", protocolStr);
    } else if (![class conformsToProtocol:serviceProtocol]) {
        exReason =  WSStr(@"Module %@ should confirm to protocol %@", moduleStr, protocolStr);
    } else {
        [self hackUnrecognizedSelecotorExceptionForModule:moduleClass];
        [WSInstance.moduleDict setObject:moduleClass forKey:protocolStr];
    }
    if (exReason.length > 0)  {
        NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];
        [userInfo setValue:@(WSExceptionFailedToRegisterModule) forKey:kWHCServieRouterExceptionCode];
        [userInfo setValue:protocolStr forKey:kWHCServieRouterExceptionServiceProtocolStr];
        NSException *exception = [[NSException alloc] initWithName:WHCServieRouterExceptionName
                                                            reason:exReason
                                                          userInfo:userInfo];
        WHCServieRouterExceptionHandler handler = [self getExceptionHandler];
        if (handler) {
            handler(exception);
        }
        WSLog(exReason);
    }
}

+ (void)unregisterService:(Protocol*_Nonnull)serviceProtocol {
    NSString *str = NSStringFromProtocol(serviceProtocol);
    if (str.length > 0) {
        [WSInstance.moduleDict removeObjectForKey:str];
    } else {
        WSLog(@"Failed to unregister service, protocol is empty");
    }
}

+ (NSArray<Class<WHCServieRouterModuleProtocol>>*_Nonnull)allRegisteredModules {
    NSArray *modules = WSInstance.moduleDict.allValues;
    NSArray *sortedModules = [modules sortedArrayUsingComparator:^NSComparisonResult(Class class1, Class class2) {
        NSUInteger priority1 = WHCServieRouterModuleDefaultPriority;
        NSUInteger priority2 = WHCServieRouterModuleDefaultPriority;
        if ([class1 respondsToSelector:@selector(priority)]) {
            priority1 = [class1 priority];
        }
        if ([class2 respondsToSelector:@selector(priority)]) {
            priority2 = [class2 priority];
        }
        if(priority1 == priority2) {
            return NSOrderedSame;
        } else if(priority1 < priority2) {
            return NSOrderedDescending;
        } else {
            return NSOrderedAscending;
        }
    }];
    return sortedModules;
}

+ (void)setupAllModules {
    NSArray *modules = [self allRegisteredModules];
    for (Class<WHCServieRouterModuleProtocol> moduleClass in modules) {
        @try {
            BOOL setupSync = NO;
            if ([moduleClass respondsToSelector:@selector(setupModuleSynchronously)]) {
                setupSync = [moduleClass setupModuleSynchronously];
            }
            if (setupSync) {
                [[moduleClass sharedInstance] setup];
            } else {
                dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
                    [[moduleClass sharedInstance] setup];
                });
            }
        } @catch (NSException *exception) {
            NSMutableDictionary *userInfo = [NSMutableDictionary dictionaryWithDictionary:exception.userInfo];
            [userInfo setValue:@(WSExceptionFailedToSetupModule) forKey:kWHCServieRouterExceptionCode];
            [userInfo setValue:NSStringFromClass(moduleClass) forKey:kWHCServieRouterExceptionModuleClassStr];
            NSException *ex = [[NSException alloc] initWithName:exception.name
                                                         reason:exception.reason
                                                       userInfo:userInfo];
            WHCServieRouterExceptionHandler handler = [self getExceptionHandler];
            if (handler) {
                 handler(ex);
            }
            WSLog(exception.reason);
        }
    }
}

+ (id<WHCServieRouterModuleProtocol> _Nullable)moduleByService:(Protocol*_Nonnull)serviceProtocol {
    NSString *protocolStr = NSStringFromProtocol(serviceProtocol);
    NSString *exReason = nil;
    NSException *exception = nil;
    if (protocolStr.length == 0) {
        exReason = WSStr(@"Invalid service protocol");
    } else {
        Class class = WSInstance.moduleDict[protocolStr];
        NSString *classStr = NSStringFromClass(class);
        if (!class) {
            exReason = WSStr(@"Failed to find module by protocol %@", protocolStr);
        } else if (![class conformsToProtocol:@protocol(WHCServieRouterModuleProtocol)]) {
            exReason = WSStr(@"Found %@ by protocol %@, but the module doesn't confirm to protocol WHCServieRouterModuleProtocol",
                            classStr, protocolStr);
        } else {
            @try {
                id instance = [class sharedInstance];
                return instance;
            } @catch (NSException *ex) {
                exception = ex;
            }
        }
    }
    if (exReason.length > 0) {
        NSExceptionName name = WHCServieRouterExceptionName;
        NSMutableDictionary *userInfo = nil;
        if (exception != nil) {
            userInfo = [NSMutableDictionary dictionaryWithDictionary:exception.userInfo];
            name = exception.name;
        } else {
            userInfo = [NSMutableDictionary dictionary];
        }
        [userInfo setValue:@(WSExceptionFailedToFindModuleByService) forKey:kWHCServieRouterExceptionCode];
        [userInfo setValue:NSStringFromProtocol(serviceProtocol) forKey:kWHCServieRouterExceptionServiceProtocolStr];
        NSException *ex = [[NSException alloc] initWithName:name
                                                            reason:exReason
                                                          userInfo:userInfo];
        WHCServieRouterExceptionHandler handler = [self getExceptionHandler];
        if (handler) {
            handler(ex);
        }
        WSLog(exReason);
        return nil;
    }
    
}

+ (BOOL)checkAllModulesWithSelector:(SEL)selector arguments:(NSArray*)arguments {
    BOOL result = NO;
    NSArray *modules = [self allRegisteredModules];
    for (Class<WHCServieRouterModuleProtocol> class in modules) {
        id<WHCServieRouterModuleProtocol> moduleItem = [class sharedInstance];
        if ([moduleItem respondsToSelector:selector]) {
            
            __block BOOL shouldInvoke = YES;
            if (![WSInstance.moduleInvokeDict objectForKey:NSStringFromClass([moduleItem class])]) {
                // 如果 modules 里面有 moduleItem 的子类，不 invoke target
                [modules enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                    if ([NSStringFromClass([obj superclass]) isEqualToString:NSStringFromClass([moduleItem class])]) {
                        shouldInvoke = NO;
                        *stop = YES;
                    }
                }];
            }
            
            if (shouldInvoke) {
                if (![WSInstance.moduleInvokeDict objectForKey:NSStringFromClass([moduleItem class])]) { //cache it
                    [WSInstance.moduleInvokeDict setObject:moduleItem forKey:NSStringFromClass([moduleItem class])];
                }
                
                BOOL ret = NO;
                [self invokeTarget:moduleItem action:selector arguments:arguments returnValue:&ret];
                if (!result) {
                    result = ret;
                }
            }
        }
    }
    return result;
}


+ (BOOL)invokeTarget:(id)target
              action:(_Nonnull SEL)selector
           arguments:(NSArray* _Nullable )arguments
         returnValue:(void* _Nullable)result; {
    if (target && [target respondsToSelector:selector]) {
        NSMethodSignature *sig = [target methodSignatureForSelector:selector];
        NSInvocation *invocation = [NSInvocation invocationWithMethodSignature:sig];
        [invocation setTarget:target];
        [invocation setSelector:selector];
        for (NSUInteger i = 0; i<[arguments count]; i++) {
            NSUInteger argIndex = i+2;
            id argument = arguments[i];
            if ([argument isKindOfClass:NSNumber.class]) {
                //convert number object to basic num type if needs
                BOOL shouldContinue = NO;
                NSNumber *num = (NSNumber*)argument;
                const char *type = [sig getArgumentTypeAtIndex:argIndex];
                if (strcmp(type, @encode(BOOL)) == 0) {
                    BOOL rawNum = [num boolValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                } else if (strcmp(type, @encode(int)) == 0
                           || strcmp(type, @encode(short)) == 0
                           || strcmp(type, @encode(long)) == 0) {
                    NSInteger rawNum = [num integerValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                } else if(strcmp(type, @encode(long long)) == 0) {
                    long long rawNum = [num longLongValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                } else if (strcmp(type, @encode(unsigned int)) == 0
                           || strcmp(type, @encode(unsigned short)) == 0
                           || strcmp(type, @encode(unsigned long)) == 0) {
                    NSUInteger rawNum = [num unsignedIntegerValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                } else if(strcmp(type, @encode(unsigned long long)) == 0) {
                    unsigned long long rawNum = [num unsignedLongLongValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                } else if (strcmp(type, @encode(float)) == 0) {
                    float rawNum = [num floatValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                } else if (strcmp(type, @encode(double)) == 0) { // double
                    double rawNum = [num doubleValue];
                    [invocation setArgument:&rawNum atIndex:argIndex];
                    shouldContinue = YES;
                }
                if (shouldContinue) {
                    continue;
                }
            }
            if ([argument isKindOfClass:[NSNull class]]) {
                argument = nil;
            }
            [invocation setArgument:&argument atIndex:argIndex];
        }
        [invocation invoke];
        NSString *methodReturnType = [NSString stringWithUTF8String:sig.methodReturnType];
        if (result && ![methodReturnType isEqualToString:@"v"]) { //if return type is not void
            if([methodReturnType isEqualToString:@"@"]) { //if it's kind of NSObject
                CFTypeRef cfResult = nil;
                [invocation getReturnValue:&cfResult]; //this operation won't retain the result
                if (cfResult) {
                    CFRetain(cfResult); //we need to retain it manually
                    *(void**)result = (__bridge_retained void *)((__bridge_transfer id)cfResult);
                }
            } else {
                [invocation getReturnValue:result];
            }
        }
        return YES;
    }
    return NO;
}

+ (void)hackUnrecognizedSelecotorExceptionForModule:(Class)class {
    SEL originSEL = @selector(doesNotRecognizeSelector:);
    SEL newSEL = @selector(bf_doesNotRecognizeSelector:);
    [self swizzleOrginSEL:originSEL withNewSEL:newSEL inClass:class];
}

+ (void)swizzleOrginSEL:(SEL)originSEL withNewSEL:(SEL)newSEL inClass:(Class)class {
    Method origMethod = class_getInstanceMethod(class, originSEL);
    Method overrideMethod = class_getInstanceMethod(class, newSEL);
    if (class_addMethod(class, originSEL, method_getImplementation(overrideMethod),
                        method_getTypeEncoding(overrideMethod))) {
        class_replaceMethod(class, newSEL, method_getImplementation(origMethod),
                            method_getTypeEncoding(origMethod));
    } else {
        method_exchangeImplementations(origMethod, overrideMethod);
    }
}

@end

@implementation NSObject (WHCServieRouter)

- (void)bf_doesNotRecognizeSelector:(SEL)aSelector {
    @try {
        [self bf_doesNotRecognizeSelector:aSelector];
    } @catch (NSException *ex) {
        NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];
        [userInfo setValue:@(WSExceptionAPINotFoundException) forKey:kWHCServieRouterExceptionCode];
        NSException *exception = [[NSException alloc] initWithName:ex.name
                                                            reason:ex.reason
                                                          userInfo:userInfo];
        if (WSInstance.exceptionHandler) {
            WSInstance.exceptionHandler(exception);
        } else {
#ifdef DEBUG
            @throw exception;
#endif
        }
    } @finally {
    }
}

@end
